# Penalized KKT MPC reformulation:

using Ipopt, JuMP

function PKKT(mod, G_0, X_0, A_0, K_g, v_gmax, N, u, G_f, sp)

    X_0 = [X_0 for i = 1:3]           # Initial consentration of biomass, gDW/L
    G_0 = [G_0 for i = 1:3]           # Initial consentration of glucose, mmol/L
    A_0 = [A_0 for i = 1:3]           # Initial consentration of acetate, mmol/L
    t =  [0.155051, 0.644949, 1.0000] # Collocation points
    i = size(mod.S)[1]                # Number of metabolites
    j = size(mod.S)[2]                # Number of reactions

    # Creating M-matrix:
    M_1 = [t[1] 1/2*t[1]^2 1/3*t[1]^3 
           t[2] 1/2*t[2]^2 1/3*t[2]^3 
           t[3] 1/2*t[3]^2 1/3*t[3]^3
    ]
    M_2 = [1 t[1] t[1]^2
           1 t[2] t[2]^2 
           1 t[3] t[3]^2
    ]
    M = M_1*M_2^(-1)

    # Creating W-matrix:
    W = zeros(Float64, j, j) 
    for k = 1:j
        for l = 1:j
            if k == l
                W[k, l] = 1e-6
            end
        end
    end

    # Initilaizing UB:
    UB = mod.UB

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-8)
    JuMP.set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[1:j, 1:N])   # Fluxes,       => Matrix: #reactions*#finite elements
    JuMP.@variable(model, G[1:3, 1:N])   # Glucose,      => Matrix: #collocation points*#finite elements
    JuMP.@variable(model, X[1:3, 1:N])   # Biomass,      => Matrix: #collocation points*#finite elements
    JuMP.@variable(model, A[1:3, 1:N])   # Acetate,      => Matrix: #collocation points*#finite elements
    JuMP.@variable(model, lam[1:i, 1:N]) # Lambda,       => Matrix: #metabolites*#finite elements
    JuMP.@variable(model, lmy[1:j, 1:N]) # LB-my,        => Matrix: #reactions*#finite elements
    JuMP.@variable(model, umy[1:j, 1:N]) # UB-my,        => Matrix: #reactions*#finite elements
    JuMP.@variable(model, LB[1:j, 1:N])  # Lower bounds  => Matrix: #reactions*#finite elements
    JuMP.@variable(model, D[1:N])        # Dilution rate => Vector: #finite elements

    # Defining LB:
    for n = 1:N
        for k = 1:20-1
            JuMP.@constraint(model, LB[k,n] == mod.LB[k])
        end
        JuMP.@NLconstraint(model, LB[20,n] == -2.5*A[3,n]/(0.01 + A[3,n]))
        for k = 20+1:28-1
            JuMP.@constraint(model, LB[k,n] == mod.LB[k])
        end
        JuMP.@NLconstraint(model, LB[28,n] == v_gmax*G[3,n]/(K_g + G[3,n]))
        for k = 28+1:j
            JuMP.@constraint(model, LB[k,n] == mod.LB[k])
        end
    end

    # Constraints for variables:
    for n = 1:N
        for k = 1:j
            JuMP.@constraint(model, v[k,n] <= UB[k])
            JuMP.@constraint(model, LB[k,n] <= v[k,n])
            JuMP.@constraint(model, lmy[k,n] >= 0)
            JuMP.@constraint(model, umy[k,n] >= 0)
        end
        for c = 1:3
            JuMP.@constraint(model, G[c,n] >= 0)
            JuMP.@constraint(model, X[c,n] >= 0)
            JuMP.@constraint(model, A[c,n] >= 0)
        end
        JuMP.@constraint(model, 0 <= D[n] <= 1)
    end
    
    # Orthogonal collocation constraints:
    JuMP.@constraint(model, X[:,1] .== X_0 .+ dt*M*(D[1]*(- X[:,1]) .+ v[13,1].*X[:,1]))
    JuMP.@constraint(model, G[:,1] .== G_0 .+ dt*M*(D[1]*(G_f .- G[:,1]) .+ v[28,1].*X[:,1]))
    JuMP.@constraint(model, A[:,1] .== A_0 .+ dt*M*(D[1]*(- A[:,1]) .+ v[20,1].*X[:,1]))
    for n = 2:N
        JuMP.@constraint(model, X[:,n] .== [X[3,n-1] for z = 1:3] .+ dt*M*(D[n]*(- X[:,n]) .+ v[13,n].*X[:,n]))
        JuMP.@constraint(model, G[:,n] .== [G[3,n-1] for z = 1:3] .+ dt*M*(D[n]*(G_f .- G[:,n]) .+ v[28,n].*X[:,n]))
        JuMP.@constraint(model, A[:,n] .== [A[3,n-1] for z = 1:3] .+ dt*M*(D[n]*(- A[:,n]) .+ v[20,n].*X[:,n]))
    end

    # Duality Constraints:
    for n = 1:N 
        JuMP.@constraint(model, mod.S*v[:,n] .== 0)
        JuMP.@constraint(model, -mod.c .+ 2 .*W*v[:,n] + transpose(mod.S)*lam[:,n] + umy[:,n] - lmy[:,n] .== 0)
    end

    # MPC constraints:
    for n in 1:N
        JuMP.@constraint(model, D[n] >= u - 0.2)
        JuMP.@constraint(model, D[n] <= u + 0.2)
    end

    # Objective:
    JuMP.@NLexpression(model, FO,  sum(sum((X[c,n] - sp)^2 for n = 1:N) for c = 1:3) + 1e-10*(u - D[1])^2 + 1e-2*sum(sum(((v[l,n] - LB[l,n])*lmy[l,n]) + (-v[l,n] + UB[l])*umy[l,n] for l in 1:j) for n in 1:N))
    JuMP.@NLobjective(model, Min, FO)

    JuMP.optimize!(model)

    return JuMP.value.(D)[1], JuMP.solve_time(model), JuMP.raw_status(model)
end
