using Ipopt, JuMP

function PKKT(mod, G_0, X_0, A_0, K_g, v_gmax)

    i = size(mod.S)[1]                # Numer of metabolites
    j = size(mod.S)[2]                # Number of reactions

    # Creating W matrix:
    W = zeros(Float64, j, j) 
    for k = 1:j
        for l = 1:j
            if k == l
                W[k, l] = 1e-6
            end
        end
    end

    # Creating UB-matrix:
    UB = mod.UB

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-10)
    set_optimizer_attribute(model, "tol", 1e-10)
    JuMP.set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[1:j])       # Fluxes
    JuMP.@variable(model, G, start=G_0) # Glucose
    JuMP.@variable(model, X, start=X_0) # Biomass
    JuMP.@variable(model, A, start=A_0) # Acetate
    JuMP.@variable(model, lam[1:i])     # Lambda
    JuMP.@variable(model, lmy[1:j])     # LB my
    JuMP.@variable(model, umy[1:j])     # UB my
    JuMP.@variable(model, LB[1:j])      # Lower bounds

    # Defining LB:
    for k = 1:20-1
        JuMP.@constraint(model, LB[k] == mod.LB[k])
    end
    JuMP.@NLconstraint(model, LB[20] == -2.5*A/(0.01 + A))
    for k = 20+1:28-1
        JuMP.@constraint(model, LB[k] == mod.LB[k])
    end
    JuMP.@NLconstraint(model, LB[28] == v_gmax*G/(K_g + G))
    for k = 28+1:j
        JuMP.@constraint(model, LB[k] == mod.LB[k])
    end

    # Constraints for variables:
    for k = 1:j
        JuMP.@constraint(model, v[k] <= UB[k])
        JuMP.@constraint(model, LB[k] <= v[k])
        JuMP.@constraint(model, lmy[k] >= 0)
        JuMP.@constraint(model, umy[k] >= 0)
    end
    JuMP.@constraint(model, G >= 0)
    JuMP.@constraint(model, X >= 1e-3)
    JuMP.@constraint(model, A >= 0)
    
    # Orthogonal collocation constraints
    JuMP.@constraint(model, 0 .== (D*(-X) + v[13]*X))
    JuMP.@constraint(model, 0 .== (D*(G_f - G) + v[28]*X))
    JuMP.@constraint(model, 0 .== (D*(-A) + v[20]*X))

    # Duality Constraints:
    JuMP.@constraint(model, mod.S*v .== 0)
    JuMP.@constraint(model, -mod.c .+ 2 .*W*v + transpose(mod.S)*lam + umy - lmy .== 0)

    JuMP.@NLexpression(model, FO, sum(((v[l] - LB[l])*lmy[l]) + (-v[l] + UB[l])*umy[l] for l in 1:j))

    # Objective:
    JuMP.@NLobjective(model, Min, FO)

    JuMP.optimize!(model)

    return JuMP.value.(G), JuMP.value.(X), JuMP.value.(A), JuMP.raw_status(model)
end