using Ipopt, JuMP

function SPKKT(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, dt)

    t_list = [0]
    G_list = [G_0]
    X_list = [X_0]
    A_list = [A_0]

    X_0 = [X_0 for i = 1:3]           # Initial consentration of biomass, g/L
    G_0 = [G_0 for i = 1:3]           # Initial consentration of glucose, mmol/L
    A_0 = [A_0 for i = 1:3] 
    t =  [0.155051, 0.644949, 1.0000] # Internal time steps
    i = size(mod.S)[1]                # Numer of metabolites
    j = size(mod.S)[2]                # Number of reactions

    # Creating M-matrix:
    M_1 = [t[1] 1/2*t[1]^2 1/3*t[1]^3 
           t[2] 1/2*t[2]^2 1/3*t[2]^3 
           t[3] 1/2*t[3]^2 1/3*t[3]^3
    ]
    M_2 = [1 t[1] t[1]^2
           1 t[2] t[2]^2 
           1 t[3] t[3]^2
    ]
    M = M_1*M_2^(-1)

    # Creating W matrix:
    W = zeros(Float64, j, j) 
    for k = 1:j
        for l = 1:j
            if k == l
                W[k, l] = 1e-6
            end
        end
    end

    # Creating UB-matrix:
    UB = reshape(vcat(mod.UB, mod.UB, mod.UB), j, 3) 

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-10)
    set_optimizer_attribute(model, "tol", 1e-10)
    JuMP.set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[1:j, 1:3, 1:N])   # Fluxes, v => Matrix: #reactions*#collocation points
    JuMP.@variable(model, G[1:3, 1:N])        # Glucose, G => Vector: #collocation points
    JuMP.@variable(model, X[1:3, 1:N])        # Biomass, X => Vector: #collocation points
    JuMP.@variable(model, A[1:3, 1:N])
    JuMP.@variable(model, lam[1:i, 1:3, 1:N]) # Lambda, lam => Matrix: #metabolites*#collocation points
    JuMP.@variable(model, lmy[1:j, 1:3, 1:N]) # LB my, lmy => Matrix: #reactions*#collocation points
    JuMP.@variable(model, umy[1:j, 1:3, 1:N]) # UB my, umy => Matrix: #reactions*#collocation points
    JuMP.@variable(model, LB[1:j, 1:3, 1:N])  # Lower bounds

    # Defining LB:
    if Ace_index < Glu_index
        for n = 1:N
            for k = 1:Ace_index-1
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
            JuMP.@NLconstraint(model, LB[Ace_index,n] == v_amax*A[3,n]/(K_a + A[3,n]))
            for k = Ace_index+1:Glu_index-1
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
            JuMP.@NLconstraint(model, LB[Glu_index,n] == v_gmax*G[3,n]/(K_g + G[3,n]))
            for k = Glu_index+1:j
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
        end
    else 
        for n = 1:N
            for k = 1:Glu_index-1
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
            JuMP.@NLconstraint(model, LB[Glu_index,n] == v_gmax*G[3,n]/(K_g + G[3,n]))
            for k = Glu_index+1:Ace_index-1
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
            JuMP.@NLconstraint(model, LB[Ace_index,n] == v_amax*A[3,n]/(K_a + A[3,n]))
            for k = Ace_index+1:j
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
        end
    end

    # Constraints for variables:
    for n = 1:N
        for c = 1:3
            for k = 1:j
                JuMP.@constraint(model, v[k,c,n] <= UB[k,c])
                JuMP.@constraint(model, LB[k,c,n] <= v[k,c,n])
                JuMP.@constraint(model, lmy[k,c,n] >= 0)
                JuMP.@constraint(model, umy[k,c,n] >= 0)
            end
            JuMP.@constraint(model, G[c,n] >= 0)
            JuMP.@constraint(model, X[c,n] >= 0)
            JuMP.@constraint(model, A[c,n] >= 0)
        end
    end
    
    #Flux constarints:
    for n = 1:N
        for k = 1:j
            JuMP.@constraint(model, v[k,2,n] - v[k,1,n] == 0)
            JuMP.@constraint(model, v[k,3,n] - v[k,2,n] == 0)
        end
    end
    
    # Orthogonal collocation constraints
    JuMP.@constraint(model, X[:,1] .== X_0 + dt*M*(v[Bio_index,1].*X[:,1]))
    JuMP.@constraint(model, G[:,1] .== G_0 + dt*M*(v[Glu_index,1].*X[:,1]))
    JuMP.@constraint(model, A[:,1] .== A_0 + dt*M*(v[Ace_index,1].*X[:,1]))
    for n = 2:N
        JuMP.@constraint(model, X[:,n] .== [X[3,n-1] for z = 1:3] + dt*M*(v[Bio_index,n].*X[:,n]))
        JuMP.@constraint(model, G[:,n] .== [G[3,n-1] for z = 1:3] + dt*M*(v[Glu_index,n].*X[:,n]))
        JuMP.@constraint(model, A[:,n] .== [A[3,n-1] for z = 1:3] + dt*M*(v[Ace_index,n].*X[:,n]))
    end

    # Duality Constraints:
    for n = 1:N 
        for c = 1:3
            JuMP.@constraint(model, mod.S*v[:,c,n] .== 0)
            JuMP.@constraint(model, -mod.c .+ 2 .*W*v[:,c,n] + transpose(mod.S)*lam[:,c,n] + umy[:,c,n] - lmy[:,c,n] .== 0)
        end
    end

    JuMP.@NLexpression(model, FO, sum(sum(sum(((v[l,c,n] - LB[l,c,n])*lmy[l,c,n]) + (-v[l,c,n] + UB[l,c])*umy[l,c,n] for l in 1:j) for c in 1:3) for n in 1:N))

    # Objective:
    JuMP.@NLobjective(model, Min, FO)

    JuMP.optimize!(model)

    # Gather the solution
    for n = 1:N
        t_list = vcat(t_list, [(n-1)*dt+0.155051*dt, (n-1)*dt+0.644949*dt, (n-1)*dt+1.0000*dt])
        for c = 1:3
            G_list = vcat(G_list, JuMP.value.(G)[c,n])
            X_list = vcat(X_list, JuMP.value.(X)[c,n])
            A_list = vcat(A_list, JuMP.value.(A)[c,n])
        end
    end

    return t_list, G_list, X_list, A_list, JuMP.raw_status(model), JuMP.value.(v)
end