using Ipopt, JuMP

function TPKKT(mod, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_0, t_end)

    t_list = [t_0]
    G_list = [G_0]
    X_list = [X_0]
    A_list = [A_0]

    X_0 = [X_0 for i = 1:3]           # Initial consentration of biomass, g/L
    G_0 = [G_0 for i = 1:3]           # Initial consentration of glucose, g/L 
    A_0 = [A_0 for i = 1:3] 
    t =  [0.155051, 0.644949, 1.0000] # Internal time steps
    i = size(mod.S)[1]                # Numer of metabolites
    j = size(mod.S)[2]                # Number of reactions

    # Creating M-matrix:
    M_1 = [t[1] 1/2*t[1]^2 1/3*t[1]^3 
           t[2] 1/2*t[2]^2 1/3*t[2]^3 
           t[3] 1/2*t[3]^2 1/3*t[3]^3
    ]
    M_2 = [1 t[1] t[1]^2
           1 t[2] t[2]^2 
           1 t[3] t[3]^2
    ]
    M = M_1*M_2^(-1)

    # Creating W matrix:
    W = zeros(Float64, j, j) 
    for k = 1:j
        for l = 1:j
            if k == l
                W[k, l] = 1e-6
            end
        end
    end

    # Creating UB-matrix:
    UB = mod.UB

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-8)
    #set_optimizer_attribute(model, "tol", 1e-10)
    JuMP.set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[1:j, 1:N])   # Fluxes, v => Matrix: #reactions*#collocation points
    JuMP.@variable(model, G[1:3, 1:N])        # Glucose, G => Vector: #collocation points
    JuMP.@variable(model, X[1:3, 1:N])        # Biomass, X => Vector: #collocation points
    JuMP.@variable(model, A[1:3, 1:N])
    JuMP.@variable(model, lam[1:i, 1:N]) # Lambda, lam => Matrix: #metabolites*#collocation points
    JuMP.@variable(model, lmy[1:j, 1:N]) # LB my, lmy => Matrix: #reactions*#collocation points
    JuMP.@variable(model, umy[1:j, 1:N]) # UB my, umy => Matrix: #reactions*#collocation points
    JuMP.@variable(model, LB[1:j, 1:N])  # Lower bounds
    JuMP.@variable(model, dt[1:N])

    # Time constraints:
    JuMP.@constraint(model, sum(dt) == t_end)
    for n = 1:N
        JuMP.@constraint(model, 0 <= dt[n] <= 2)
    end

    # Defining LB:
    if Ace_index < Glu_index
        for n = 1:N
            for k = 1:Ace_index-1
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
            JuMP.@NLconstraint(model, LB[Ace_index,n] == v_amax*A[3,n]/(K_a + A[3,n]))
            for k = Ace_index+1:Glu_index-1
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
            JuMP.@NLconstraint(model, LB[Glu_index,n] == v_gmax*G[3,n]/(K_g + G[3,n]))
            for k = Glu_index+1:j
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
        end
    else 
        for n = 1:N
            for k = 1:Glu_index-1
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
            JuMP.@NLconstraint(model, LB[Glu_index,n] == v_gmax*G[3,n]/(K_g + G[3,n]))
            for k = Glu_index+1:Ace_index-1
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
            JuMP.@NLconstraint(model, LB[Ace_index,n] == v_amax*A[3,n]/(K_a + A[3,n]))
            for k = Ace_index+1:j
                JuMP.@constraint(model, LB[k,n] == mod.LB[k])
            end
        end
    end

    # Constraints for variables:
    for n = 1:N
        for k = 1:j
            JuMP.@constraint(model, v[k,n] <= UB[k])
            JuMP.@constraint(model, LB[k,n] <= v[k,n])
            JuMP.@constraint(model, lmy[k,n] >= 0)
            JuMP.@constraint(model, umy[k,n] >= 0)
        end
        for c = 1:3
            JuMP.@constraint(model, G[c,n] >= 0)
            JuMP.@constraint(model, X[c,n] >= 0)
            JuMP.@constraint(model, A[c,n] >= 0)
        end
    end
    
    # Orthogonal collocation constraints
    for c = 1:3
        JuMP.@NLconstraint(model, X[c,1] == X_0[c] + dt[1]*(M[c,1]*((v[Bio_index,1]-D)*X[1,1]) + M[c,2]*((v[Bio_index,1]-D)*X[2,1]) + M[c,3]*((v[Bio_index,1]-D)*X[3,1])))
        JuMP.@NLconstraint(model, G[c,1] == G_0[c] + dt[1]*(M[c,1]*(D*(G_f - G[1,1])+v[Glu_index,1]*X[1,1]) + M[c,2]*(D*(G_f - G[2,1])+v[Glu_index,1]*X[2,1]) + M[c,3]*(D*(G_f - G[3,1])+v[Glu_index,1]*X[3,1])))
        JuMP.@NLconstraint(model, A[c,1] == A_0[c] + dt[1]*(M[c,1]*(D*(- A[1,1])+v[Ace_index,1]*X[1,1]) + M[c,2]*(D*(- A[2,1])+v[Ace_index,1]*X[2,1]) + M[c,3]*(D*(- A[3,1])+v[Ace_index,1]*X[3,1])))
        for n = 2:N
            JuMP.@NLconstraint(model, X[c,n] == X[3,n-1] + dt[n]*(M[c,1]*((v[Bio_index,n]-D)*X[1,n]) + M[c,2]*((v[Bio_index,n]-D)*X[2,n]) + M[c,3]*((v[Bio_index,n]-D)*X[3,n])))
            JuMP.@NLconstraint(model, G[c,n] == G[3,n-1] + dt[n]*(M[c,1]*(D*(G_f - G[1,n])+v[Glu_index,n]*X[1,n]) + M[c,2]*(D*(G_f - G[2,n])+v[Glu_index,n]*X[2,n]) + M[c,3]*(D*(G_f - G[3,n])+v[Glu_index,n]*X[3,n])))
            JuMP.@NLconstraint(model, A[c,n] == A[3,n-1] + dt[n]*(M[c,1]*(D*(- A[1,n])+v[Ace_index,n]*X[1,n]) + M[c,2]*(D*(- A[2,n])+v[Ace_index,n]*X[2,n]) + M[c,3]*(D*(- A[3,n])+v[Ace_index,n]*X[3,n])))
        end
    end

    # Duality Constraints:
    for n = 1:N 
        JuMP.@constraint(model, mod.S*v[:,n] .== 0)
        JuMP.@constraint(model, -mod.c .+ 2 .*W*v[:,n] .+ transpose(mod.S)*lam[:,n] .+ umy[:,n] .- lmy[:,n] .== 0)
    end

    JuMP.@NLexpression(model, FO, sum(sum(((v[l,n] - LB[l,n])*lmy[l,n]) + (-v[l,n] + UB[l])*umy[l,n] for l in 1:j) for n in 1:N) + 0.01*sum(dt[n]*(G[1,n]-G[3,n])^2 for n=1:N))

    # Objective:
    JuMP.@NLobjective(model, Min, FO)

    JuMP.optimize!(model)

    # Gather the solution
    for n = 1:N
        t_list = vcat(t_list, [last(t_list)+0.155051*JuMP.value.(dt)[n], last(t_list)+0.644949*JuMP.value.(dt)[n], last(t_list)+1.0000*JuMP.value.(dt)[n]])
        for c = 1:3
            G_list = vcat(G_list, JuMP.value.(G)[c,n])
            X_list = vcat(X_list, JuMP.value.(X)[c,n])
            A_list = vcat(A_list, JuMP.value.(A)[c,n])
        end
    end

    return t_list, G_list, X_list, A_list, JuMP.raw_status(model)
end
