using Ipopt, DifferentialEquations, JuMP

# pFBA solver:
function pFBA(mod)

    S = mod.S
    c = mod.c
    UB = mod.UB
    LB = mod.LB

    W = zeros(Float64, size(S)[2], size(S)[2])
    for i = 1:size(S)[2]
        for j = 1:size(S)[2]
            if i == j
                W[i, j] = 1e-6
            end
        end
    end   

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    JuMP.set_silent(model)

    JuMP.@variable(model, v[i=1:length(c)], start = mod.v[i])

    for i = 1:length(c)
        JuMP.@constraint(model, LB[i] <= v[i] <= UB[i])
    end

    JuMP.@constraint(model, S*v .== 0)

    JuMP.@expression(model, FO, -transpose(c)*v + transpose(v)*W*v)

    JuMP.@objective(model, Min, FO)

    JuMP.optimize!(model)

    v_opt = JuMP.value.(v)

    for i = 1:length(c)
        mod.v[i] = v_opt[i] 
    end

    return nothing
end

# Reactor rquations:
function reactor!(dC, C, consts, t)

    # Gathering constants:
    # ----------------------------------------------------
    K_g       = consts[1] # g/L
    K_a       = consts[2] # g/L
    v_gmax    = consts[3] # mol/gDW h
    v_amax    = consts[4] # mol/gDW h
    model     = consts[5] # -
    Bio_index = consts[6] # -
    Ace_index = consts[7] # -
    Glu_index = consts[8] # -

    # Gathering consentrations
    X = C[1]
    G = C[2]
    A = C[3]

    if A < 0
        A = 0
        changeRxnBounds(model, "EX_ac(e)", 0, "l")
    else
        v_a = v_amax*(A/(K_a + A))
        changeRxnBounds(model, "EX_ac(e)", v_a, "l")
    end
    if G < 0
        G = 0
        changeRxnBounds(model, "EX_glc(e)", 0, "l")
    else
        v_g = v_gmax*(G/(K_g + G))
        changeRxnBounds(model, "EX_glc(e)", v_g, "l")
    end

    pFBA(model)
    mu, v_g, v_a = model.v[Bio_index], model.v[Glu_index], model.v[Ace_index]

    # Mass balance equations:
    # ----------------------------------------------------
    dXdt = -D*X + mu*X
    dGdt = D*(G_f - G) + v_g*X
    dAdt = -D.*A + v_a*X

    # Return result:
    # ----------------------------------------------------
    dC .= vcat(dXdt, dGdt, dAdt)
    
    return nothing
end

# ---
function DAPsave(model, G_0, X_0, A_0, K_g, K_a, v_gmax, v_amax, Bio_index, Ace_index, Glu_index, t_end, save_at)
    # Initialization: 
    # ----------------------------------------------------
    init = vcat(X_0, G_0, A_0)
    consts = vcat(K_g, K_a, v_gmax, v_amax, model, Bio_index, Ace_index, Glu_index)

    # Initializing boundary in t-direction:
    #-------------------------------------------------------------------
    tspan = [0, t_end] # s
    
    # Solving the ode-system:
    #-------------------------------------------------------------------
    prob = ODEProblem(reactor!, init, tspan, consts)
    sol = DifferentialEquations.solve(prob, QNDF(autodiff=false), reltol=1e-6, abstol=1e-6, saveat=save_at)

    display(sol.destats)

        # Gathering results:
    #-------------------------------------------------------------------
    t = Array{Float64}(undef,size(sol.t,1))
    C_X = Array{Float64}(undef,size(sol.t,1))
    C_G = Array{Float64}(undef,size(sol.t,1))
    C_A = Array{Float64}(undef,size(sol.t,1))

    # t2 - tend:
    #-------------------------------------------------------------------
    for i=1:size(sol.t,1)
        t[i] = sol.t[i]
        C_X[i] = sol[i][1,1]
        C_G[i] = sol[i][2,1]
        C_A[i] = sol[i][3,1]
    end

    # Reset model
    for i = 1:length(mod.v)
        mod.v[i] = 0.0 
    end

    return t, C_G, C_X, C_A
end