using JuMP, Ipopt

# org. pFBA:
function pFBA(mod)

    S = mod.S
    c = mod.c
    UB = mod.UB
    LB = mod.LB

    W = zeros(Float64, size(S)[2], size(S)[2])
    for i = 1:size(S)[2]
        for j = 1:size(S)[2]
            if i == j
                W[i, j] = 1e-6
            end
        end
    end   

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-8)
    #set_optimizer_attribute(model, "tol", 1e-10)
    set_silent(model)

    JuMP.@variable(model, v[i=1:length(c)])

    for i = 1:length(c)
        JuMP.@constraint(model, LB[i] <= v[i] <= UB[i])
    end

    JuMP.@constraint(model, S*v .== 0)

    JuMP.@expression(model, FO, -transpose(c)*v + transpose(v)*W*v)

    JuMP.@objective(model, Min, FO)

    JuMP.optimize!(model)

    return JuMP.value.(v), JuMP.solve_time(model), JuMP.raw_status(model)
end


# pFBA - Dual:
function Dual(mod)

    S = mod.S
    c = mod.c
    UB = mod.UB
    LB = mod.LB

    i = size(S)[1]
    j = size(S)[2]
    W = zeros(Float64, size(S)[2], size(S)[2])
    for i = 1:size(S)[2]
        for j = 1:size(S)[2]
            if i == j
                W[i, j] = 1e-6
            end
        end
    end   

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-8)
    #set_optimizer_attribute(model, "tol", 1e-10)
    set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[k = 1:j]) # Fluxes, x
    JuMP.@variable(model, lam[k = 1:i]) # Lambda, lam
    JuMP.@variable(model, lmy[k = 1:j]) # LB my, lmy
    JuMP.@variable(model, umy[k = 1:j]) # UB my, umy

    # Constraints for variables:
    for k = 1:j
        JuMP.@constraint(model, LB[k] <= v[k] <= UB[k])
        JuMP.@constraint(model, lmy[k] >= 0)
        JuMP.@constraint(model, umy[k] >= 0)
    end
    
    # Constraints:
    JuMP.@constraint(model, S*v .== 0)

    JuMP.@constraint(model, -c .+ 2 .*W*v .+ transpose(S)*lam .+ umy .- lmy .== 0)

    JuMP.@constraint(model, -transpose(v)*W*v + transpose(lmy)*LB - transpose(umy)*UB == -transpose(c)*v + transpose(v)*W*v)

    # Objective:
    JuMP.@objective(model, Min, 1)

    JuMP.optimize!(model)

    return JuMP.value.(v), JuMP.solve_time(model), JuMP.raw_status(model)
end

# pFBA - P. Dual:
function PDual(mod)

    S = mod.S
    c = mod.c
    UB = mod.UB
    LB = mod.LB

    i = size(S)[1]
    j = size(S)[2]
    W = zeros(Float64, size(S)[2], size(S)[2])
    for i = 1:size(S)[2]
        for j = 1:size(S)[2]
            if i == j
                W[i, j] = 1e-6
            end
        end
    end   

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-8)
    #set_optimizer_attribute(model, "tol", 1e-10)
    set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[k = 1:j]) # Fluxes, x
    JuMP.@variable(model, lam[k = 1:i]) # Lambda, lam
    JuMP.@variable(model, lmy[k = 1:j]) # LB my, lmy
    JuMP.@variable(model, umy[k = 1:j]) # UB my, umy

    # Constraints for variables:
    for k = 1:j
        JuMP.@constraint(model, LB[k] <= v[k] <= UB[k])
        JuMP.@constraint(model, lmy[k] >= 0)
        JuMP.@constraint(model, umy[k] >= 0)
    end
    
    # Constraints:
    JuMP.@constraint(model, S*v .== 0)

    JuMP.@constraint(model, -c .+ 2 .*W*v .+ transpose(S)*lam .+ umy .- lmy .== 0)

    JuMP.@NLexpression(model, FO, (sum(2*v[k]*W[k,k]*v[k] - LB[k]*lmy[k] + UB[k]*umy[k] - c[k]*v[k] for k =1:j)))

    # Objective:
    JuMP.@NLobjective(model, Min, FO)

    JuMP.optimize!(model)

    return JuMP.value.(v), JuMP.solve_time(model), JuMP.raw_status(model)
end

# pFBA - P. Dual squared:
function PDual2(mod)

    S = mod.S
    c = mod.c
    UB = mod.UB
    LB = mod.LB

    i = size(S)[1]
    j = size(S)[2]
    W = zeros(Float64, size(S)[2], size(S)[2])
    for i = 1:size(S)[2]
        for j = 1:size(S)[2]
            if i == j
                W[i, j] = 1e-6
            end
        end
    end   

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-8)
    #set_optimizer_attribute(model, "tol", 1e-10)
    set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[k = 1:j]) # Fluxes, x
    JuMP.@variable(model, lam[k = 1:i]) # Lambda, lam
    JuMP.@variable(model, lmy[k = 1:j]) # LB my, lmy
    JuMP.@variable(model, umy[k = 1:j]) # UB my, umy

    # Constraints for variables:
    for k = 1:j
        JuMP.@constraint(model, LB[k] <= v[k] <= UB[k])
        JuMP.@constraint(model, lmy[k] >= 0)
        JuMP.@constraint(model, umy[k] >= 0)
    end
    
    # Constraints:
    JuMP.@constraint(model, S*v .== 0)

    JuMP.@constraint(model, -c .+ 2 .*W*v .+ transpose(S)*lam .+ umy .- lmy .== 0)

    JuMP.@NLexpression(model, FO, (sum(2*v[k]*W[k,k]*v[k] - LB[k]*lmy[k] + UB[k]*umy[k] - c[k]*v[k] for k =1:j))^2)

    # Objective:
    JuMP.@NLobjective(model, Min, FO)

    JuMP.optimize!(model)

    return JuMP.value.(v), JuMP.solve_time(model), JuMP.raw_status(model)
end

# pFBA - KKT:
function KKT(mod)

    S = mod.S
    c = mod.c
    UB = mod.UB
    LB = mod.LB

    i = size(S)[1]
    j = size(S)[2]
    W = zeros(Float64, size(S)[2], size(S)[2])
    for i = 1:size(S)[2]
        for j = 1:size(S)[2]
            if i == j
                W[i, j] = 1e-6
            end
        end
    end   

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-8)
    #set_optimizer_attribute(model, "tol", 1e-10)
    set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[k = 1:j]) # Fluxes, x
    JuMP.@variable(model, lam[k = 1:i]) # Lambda, lam
    JuMP.@variable(model, lmy[k = 1:j]) # LB my, lmy
    JuMP.@variable(model, umy[k = 1:j]) # UB my, umy

    # Constraints for variables:
    for k = 1:j
        JuMP.@constraint(model, LB[k] <= v[k] <= UB[k])
        JuMP.@constraint(model, lmy[k] >= 0)
        JuMP.@constraint(model, umy[k] >= 0)
    end
    
    # Constraints:
    JuMP.@constraint(model, S*v .== 0)

    JuMP.@constraint(model, -c .+ 2 .*W*v .+ transpose(S)*lam .+ umy .- lmy .== 0)

    JuMP.@constraint(model, transpose(lmy)*(v .- LB) == 0)
    JuMP.@constraint(model, transpose(umy)*(UB .- v) == 0)

    # Objective:
    JuMP.@objective(model, Min, 1)

    JuMP.optimize!(model)

    return JuMP.value.(v), JuMP.solve_time(model), JuMP.raw_status(model)
end

# pFBA - P. KKT:
function PKKT(mod)

    S = mod.S
    c = mod.c
    UB = mod.UB
    LB = mod.LB

    i = size(S)[1]
    j = size(S)[2]
    W = zeros(Float64, size(S)[2], size(S)[2])
    for i = 1:size(S)[2]
        for j = 1:size(S)[2]
            if i == j
                W[i, j] = 1e-6
            end
        end
    end   

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-8)
    #set_optimizer_attribute(model, "tol", 1e-10)
    set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[k = 1:j]) # Fluxes, x
    JuMP.@variable(model, lam[k = 1:i]) # Lambda, lam
    JuMP.@variable(model, lmy[k = 1:j]) # LB my, lmy
    JuMP.@variable(model, umy[k = 1:j]) # UB my, umy

    # Constraints for variables:
    for k = 1:j
        JuMP.@constraint(model, LB[k] <= v[k] <= UB[k])
        JuMP.@constraint(model, lmy[k] >= 0)
        JuMP.@constraint(model, umy[k] >= 0)
    end
    
    # Constraints:
    JuMP.@constraint(model, S*v .== 0)

    JuMP.@constraint(model, -c .+ 2 .*W*v .+ transpose(S)*lam .+ umy .- lmy .== 0)

    JuMP.@expression(model, FO, transpose(lmy)*(v .- LB) + transpose(umy)*(UB .- v))

    # Objective:
    JuMP.@objective(model, Min, FO)

    JuMP.optimize!(model)

    return JuMP.value.(v), JuMP.solve_time(model), JuMP.raw_status(model)
end

# pFBA - P. KKT squared:
function PKKT2(mod)

    S = mod.S
    c = mod.c
    UB = mod.UB
    LB = mod.LB

    i = size(S)[1]
    j = size(S)[2]
    W = zeros(Float64, size(S)[2], size(S)[2])
    for i = 1:size(S)[2]
        for j = 1:size(S)[2]
            if i == j
                W[i, j] = 1e-6
            end
        end
    end   

    model = Model(Ipopt.Optimizer)
    set_optimizer_attribute(model, "linear_solver", "ma97")
    set_optimizer_attribute(model, "constr_viol_tol", 1e-10)
    set_optimizer_attribute(model, "acceptable_tol", 1e-8)
    #set_optimizer_attribute(model, "tol", 1e-10)
    set_silent(model)

    # Variabels:
    JuMP.@variable(model, v[k = 1:j]) # Fluxes, x
    JuMP.@variable(model, lam[k = 1:i]) # Lambda, lam
    JuMP.@variable(model, lmy[k = 1:j]) # LB my, lmy
    JuMP.@variable(model, umy[k = 1:j]) # UB my, umy

    # Constraints for variables:
    for k = 1:j
        JuMP.@constraint(model, LB[k] <= v[k] <= UB[k])
        JuMP.@constraint(model, lmy[k] >= 0)
        JuMP.@constraint(model, umy[k] >= 0)
    end
    
    # Constraints:
    JuMP.@constraint(model, S*v .== 0)

    JuMP.@constraint(model, -c .+ 2 .*W*v .+ transpose(S)*lam .+ umy .- lmy .== 0)

    JuMP.@NLexpression(model, FO, sum(lmy[k]*(v[k] - LB[k]) + umy[k]*(UB[k] - v[k]) for k=1:j)^2)

    # Objective:
    JuMP.@NLobjective(model, Min, FO)

    JuMP.optimize!(model)

    return JuMP.value.(v), JuMP.solve_time(model), JuMP.raw_status(model)
end
